package Algorithm.parMCE;
/**
 * ParVertexParMCE : running subproblems sequentially using ParTTT
 * */

import java.util.*;

import Algorithm.Graph;
import utils.MurmurHash3;
import utils.SetOperations;

import java.io.*;
import java.util.concurrent.atomic.*;
import java.util.stream.IntStream;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ForkJoinPool;

public class ParVertexParMCE_Total {

	// private Set<Integer> CAND;
	// private Set<Integer> FINI;
	// private TreeSet<Integer> K;
	// private ArrayList<Integer> K;
	// public HashSet<TreeSet<Integer>> CLQ;
	private static Collection<BitSet> CLQ;

	private static ConcurrentMap<Long, Integer> timeToFrequency;
	public int clqcnt;
	private FileWriter cw;

	private static Graph G;

	static AtomicInteger count;

	private static HashMap<Integer, HashSet<Integer>> MapVE;
	
	static long loopoverhead;

	public ParVertexParMCE_Total(Graph g, String ofname) throws IOException {

		this.G = g;

		count = new AtomicInteger(0);
		
		loopoverhead = 0;

		CLQ = ConcurrentHashMap.newKeySet();

		timeToFrequency = new ConcurrentHashMap<>();
		
		long elapsed = 0L;
		int numOfVertices = G.numV() ;
		for (int v = 0 ; v < numOfVertices ; v ++) {
			Set<Integer> CAND = new HashSet<Integer>();
			Set<Integer> FINI = new HashSet<Integer>();
			TreeSet<Integer> K = new TreeSet<Integer>();

			K.add(v);
			for (int w : G.AdjList.get(v)) {
				if (G.degreeOf(w) > G.degreeOf(v))
					CAND.add(w);
				else if (G.degreeOf(w) < G.degreeOf(v))
					FINI.add(w);
				else {
					if (w > v)
						CAND.add(w);
					else
						FINI.add(w);
				}
			}
			try {
				long t1 = System.currentTimeMillis();
				int cand_size = CAND.size();
				int fini_size = FINI.size();
				expand(K, CAND, FINI);
				elapsed = System.currentTimeMillis() - t1;
				System.out.println(v + "\t" + elapsed + "\t" + cand_size + "\t" + fini_size);

			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
	}

	public void expand(TreeSet<Integer> K, Collection<Integer> CAND, Collection<Integer> FINI) throws IOException {
		if (CAND.isEmpty() && FINI.isEmpty()) {
			
			count.incrementAndGet();
			//if (count.get() % 100000000 == 0)
			//	System.out.println(count.get());

			return;
		}
		int u = find_u(CAND, FINI);

		Collection<Integer> NghOfu = Ngh(G, u);

		List<Integer> Ext = new ArrayList<>();


		for(int w : CAND){
			if (!NghOfu.contains(w)) {
				Ext.add(w);
			}
		}

		int size = Ext.size();

		
		if (size > 0) {
			IntStream.range(0, size).parallel().forEach(i -> {
				
				//long start = System.currentTimeMillis();

				int q = Ext.get(i);
				TreeSet<Integer> Kq = new TreeSet<>();
				Kq.addAll(K);
				Kq.add(q);

				Collection<Integer> NghOfq = Ngh(G, q);

				HashSet<Integer> CANDq = SetOperations.intersect(CAND, NghOfq);
				//Collection<Integer> CANDq = SetOperations.parIntersect(CAND, NghOfq);

				HashSet<Integer> FINIq = SetOperations.intersect(FINI, NghOfq);
				//Collection<Integer> FINIq = SetOperations.parIntersect(FINI, NghOfq);

				for (int j = 0; j < i; j++) {
					int p = Ext.get(j);
					CANDq.remove(p);
					if (NghOfq.contains(p))
						FINIq.add(p);
				}
				
				//loopoverhead += (System.currentTimeMillis() - start);

				try {
					expand(Kq, CANDq, FINIq);
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

			});
		}
	}

	/*
	 * public Set<Integer> Ngh(int u) { return G.AdjList.get(u); }
	 */

	public Collection<Integer> Ngh(Graph G, int u) {
		if (G.AdjList.get(u) == null)
			System.out.println("No neighbors");
		return G.AdjList.get(u);
	}

	public int find_u(Collection<Integer> CAND, Collection<Integer> FINI /*
															 * Set<String> SUBG
															 */) {

		Map<Integer, Integer> vToIntersectSize = new ConcurrentHashMap<>(); 
		
		CAND.parallelStream().forEach(u -> {
			
			HashSet<Integer> Q = SetOperations.intersect(Ngh(G,u), CAND);
			//Collection<Integer> Q = SetOperations.parIntersect(Ngh(G,u), CAND);
			vToIntersectSize.put(u, Q.size());
			
		});
		
		FINI.parallelStream().forEach(u -> {
			
			HashSet<Integer> Q = SetOperations.intersect(Ngh(G,u), CAND);
			//Collection<Integer> Q = SetOperations.parIntersect(Ngh(G,u), CAND);
			vToIntersectSize.put(u, Q.size());
			
		});
		
		int size = -1;
		int v = 0;
		for(int u : vToIntersectSize.keySet()) {
			int tmp = vToIntersectSize.get(u);
			if(size < tmp) {
				size = tmp;
				v = u;
			}
		}

		return v;

	}

	public static void main(String[] args) {
		// TODO Auto-generated method stub

		System.setProperty("java.util.concurrent.ForkJoinPool.common.parallelism", args[1]);
		
		System.out.println("ParVertexParMCE : running subproblems sequentially using ParTTT");
		System.out.println("Input Graph: " + args[0]);

		Graph G = new Graph(args[0]);

		System.out.println("Graph Reading Complete");

		try {
			long t1 = System.currentTimeMillis();
			new ParVertexParMCE_Total(G, "output_TTT");
			long elapsed = System.currentTimeMillis() - t1;
			//System.out.println("number of maximal cliques: " + CLQ.size());
			System.out.println("number of maximal cliques: " + count.get());
			System.out.println(
					"Time taken to compute maximal cliques in " + args[0] + " : " + elapsed / 1000. + " sec.");
			//System.out.println("Total loop overhead : " + loopoverhead/1000 + " sec.");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// new Tomita(G);
	}

}
